From 315a5af8f25fbb6aabcbe50a9656c198c581f881 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Fri, 25 Jun 2021 15:52:31 +0000
Subject: [PATCH] [arm64] port LUMA_VPP_4xN

luma_vpp[  4x4]		18.77x 	 27.66    	 519.22
luma_vpp[  4x8]		22.73x 	 45.35    	 1030.72
luma_vpp[ 4x16]		25.10x 	 82.32    	 2066.41
---
 source/common/aarch64/asm-primitives.cpp |  5 ++
 source/common/aarch64/asm.S              | 43 +++++++++++
 source/common/aarch64/ipfilter8.S        | 94 ++++++++++++++++++++++++
 source/common/aarch64/ipfilter8.h        |  4 +
 source/common/arm/ipfilter8.S            |  2 +-
 5 files changed, 147 insertions(+), 1 deletion(-)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index 592d183..76c9882 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -240,6 +240,11 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.pu[LUMA_64x32].convert_p2s[NONALIGNED] = PFX(filterPixelToShort_64x32_neon);
         p.pu[LUMA_64x48].convert_p2s[NONALIGNED] = PFX(filterPixelToShort_64x48_neon);
         p.pu[LUMA_64x64].convert_p2s[NONALIGNED] = PFX(filterPixelToShort_64x64_neon);
+
+        // vertical interpolation filters
+        p.pu[LUMA_4x4].luma_vpp     = PFX(interp_8tap_vert_pp_4x4_neon);
+        p.pu[LUMA_4x8].luma_vpp     = PFX(interp_8tap_vert_pp_4x8_neon);
+        p.pu[LUMA_4x16].luma_vpp    = PFX(interp_8tap_vert_pp_4x16_neon);
     }
 }
 } // namespace X265_NS
diff --git a/source/common/aarch64/asm.S b/source/common/aarch64/asm.S
index 5f020a1..88f81da 100644
--- a/source/common/aarch64/asm.S
+++ b/source/common/aarch64/asm.S
@@ -35,6 +35,12 @@
 #define ELF @
 #endif
 
+#ifdef __MACH__
+#   define MACH
+#else
+#   define MACH #
+#endif
+
 #define HAVE_AS_FUNC 1
 
 #if HAVE_AS_FUNC
@@ -64,6 +70,43 @@ FUNC    .func   \name
 .endif
 .endm
 
+.macro  const   name, align=2
+    .macro endconst
+ELF     .size   \name, . - \name
+        .purgem endconst
+    .endm
+ELF     .section        .rodata
+MACH    .const_data
+        .align          \align
+\name:
+.endm
+
+.macro  movrel rd, val, offset=0
+#if defined(__APPLE__)
+  .if \offset < 0
+        adrp            \rd, \val@PAGE
+        add             \rd, \rd, \val@PAGEOFF
+        sub             \rd, \rd, -(\offset)
+  .else
+        adrp            \rd, \val+(\offset)@PAGE
+        add             \rd, \rd, \val+(\offset)@PAGEOFF
+  .endif
+#elif defined(PIC) && defined(_WIN32)
+  .if \offset < 0
+        adrp            \rd, \val
+        add             \rd, \rd, :lo12:\val
+        sub             \rd, \rd, -(\offset)
+  .else
+        adrp            \rd, \val+(\offset)
+        add             \rd, \rd, :lo12:\val+(\offset)
+  .endif
+#elif defined(PIC)
+        adrp            \rd, \val+(\offset)
+        add             \rd, \rd, :lo12:\val+(\offset)
+#else
+        ldr             \rd, =\val+\offset
+#endif
+.endm
 
 #define FENC_STRIDE 64
 #define FDEC_STRIDE 32
diff --git a/source/common/aarch64/ipfilter8.S b/source/common/aarch64/ipfilter8.S
index d467952..5d412f7 100644
--- a/source/common/aarch64/ipfilter8.S
+++ b/source/common/aarch64/ipfilter8.S
@@ -240,6 +240,100 @@ function x265_filterPixelToShort_48x64_neon
     ret
 endfunc
 
+//**************luma_vpp************
+
+const g_luma_s16, align=8
+.hword   0, 0,   0, 64,  0,   0, 0,  0
+.hword  -1, 4, -10, 58, 17,  -5, 1,  0
+.hword  -1, 4, -11, 40, 40, -11, 4, -1
+.hword   0, 1,  -5, 17, 58, -10, 4, -1
+endconst
+
+.macro LUMA_VPP_4xN h
+function x265_interp_8tap_vert_pp_4x\h\()_neon
+    movrel          x10, g_luma_s16
+    sub             x0, x0, x1
+    sub             x0, x0, x1, lsl #1         // src -= 3 * srcStride
+    lsl             x4, x4, #4
+    ldr             q0, [x10, x4]              // q0 = luma interpolate coeff
+    dup             v24.8h, v0.h[0]
+    dup             v25.8h, v0.h[1]
+    trn1            v24.2d, v24.2d, v25.2d
+    dup             v26.8h, v0.h[2]
+    dup             v27.8h, v0.h[3]
+    trn1            v26.2d, v26.2d, v27.2d
+    dup             v28.8h, v0.h[4]
+    dup             v29.8h, v0.h[5]
+    trn1            v28.2d, v28.2d, v29.2d
+    dup             v30.8h, v0.h[6]
+    dup             v31.8h, v0.h[7]
+    trn1            v30.2d, v30.2d, v31.2d
+
+    // prepare to load 8 lines
+    ld1             {v0.s}[0], [x0], x1
+    ld1             {v0.s}[1], [x0], x1
+    ushll           v0.8h, v0.8b, #0
+    ld1             {v1.s}[0], [x0], x1
+    ld1             {v1.s}[1], [x0], x1
+    ushll           v1.8h, v1.8b, #0
+    ld1             {v2.s}[0], [x0], x1
+    ld1             {v2.s}[1], [x0], x1
+    ushll           v2.8h, v2.8b, #0
+    ld1             {v3.s}[0], [x0], x1
+    ld1             {v3.s}[1], [x0], x1
+    ushll           v3.8h, v3.8b, #0
+
+    mov             x9, #\h
+.loop_4x\h:
+    ld1             {v4.s}[0], [x0], x1
+    ld1             {v4.s}[1], [x0], x1
+    ushll           v4.8h, v4.8b, #0
+
+    // row[0-1]
+    mul             v16.8h, v0.8h, v24.8h
+    ext             v21.16b, v0.16b, v1.16b, #8
+    mul             v17.8h, v21.8h, v24.8h
+    orr             v0.16b, v1.16b, v1.16b
+
+    // row[2-3]
+    mla             v16.8h, v1.8h, v26.8h
+    ext             v21.16b, v1.16b, v2.16b, #8
+    mla             v17.8h, v21.8h, v26.8h
+    orr             v1.16b, v2.16b, v2.16b
+
+    // row[4-5]
+    mla             v16.8h, v2.8h, v28.8h
+    ext             v21.16b, v2.16b, v3.16b, #8
+    mla             v17.8h, v21.8h, v28.8h
+    orr             v2.16b, v3.16b, v3.16b
+
+    // row[6-7]
+    mla             v16.8h, v3.8h, v30.8h
+    ext             v21.16b, v3.16b, v4.16b, #8
+    mla             v17.8h, v21.8h, v30.8h
+    orr             v3.16b, v4.16b, v4.16b
+
+    // sum row[0-7]
+    dup             v18.2d, v16.d[1]
+    dup             v19.2d, v17.d[1]
+    add             v16.4h, v16.4h, v18.4h
+    add             v17.4h, v17.4h, v19.4h
+
+    trn1            v16.2d, v16.2d, v17.2d
+    sqrshrun        v16.8b,  v16.8h,  #6
+    st1             {v16.s}[0], [x2], x3
+    st1             {v16.s}[1], [x2], x3
+
+    sub             x9, x9, #2
+    cbnz            x9, .loop_4x\h
+    ret
+endfunc
+.endm
+
+LUMA_VPP_4xN 4
+LUMA_VPP_4xN 8
+LUMA_VPP_4xN 16
+
 .macro qpel_filter_0_32b
     movi            v24.8h, #64
     uxtl            v19.8h, v5.8b
diff --git a/source/common/aarch64/ipfilter8.h b/source/common/aarch64/ipfilter8.h
index 2bc81bd..f3324ec 100644
--- a/source/common/aarch64/ipfilter8.h
+++ b/source/common/aarch64/ipfilter8.h
@@ -50,6 +50,10 @@ void x265_filterPixelToShort_64x32_neon(const pixel* src, intptr_t srcStride, in
 void x265_filterPixelToShort_64x48_neon(const pixel* src, intptr_t srcStride, int16_t* dst, intptr_t dstStride);
 void x265_filterPixelToShort_64x64_neon(const pixel* src, intptr_t srcStride, int16_t* dst, intptr_t dstStride);
 
+void x265_interp_8tap_vert_pp_4x4_neon(const pixel* src, intptr_t srcStride, pixel* dst, intptr_t dstStride, int coeffIdx);
+void x265_interp_8tap_vert_pp_4x8_neon(const pixel* src, intptr_t srcStride, pixel* dst, intptr_t dstStride, int coeffIdx);
+void x265_interp_8tap_vert_pp_4x16_neon(const pixel* src, intptr_t srcStride, pixel* dst, intptr_t dstStride, int coeffIdx);
+
 void x265_interp_8tap_horiz_ps_4x4_neon(const pixel* src, intptr_t srcStride, int16_t* dst, intptr_t dstStride, int coeffIdx, int isRowExt);
 void x265_interp_8tap_horiz_ps_4x8_neon(const pixel* src, intptr_t srcStride, int16_t* dst, intptr_t dstStride, int coeffIdx, int isRowExt);
 void x265_interp_8tap_horiz_ps_4x16_neon(const pixel* src, intptr_t srcStride, int16_t* dst, intptr_t dstStride, int coeffIdx, int isRowExt);
diff --git a/source/common/arm/ipfilter8.S b/source/common/arm/ipfilter8.S
index 8b7f5b3..c6ac166 100644
--- a/source/common/arm/ipfilter8.S
+++ b/source/common/arm/ipfilter8.S
@@ -795,7 +795,7 @@ function x265_interp_8tap_vert_pp_4x\h\()_neon
     vst1.u32    {d18[1]}, [r2], r3
 
     subs        r12, #2
-    bne        .loop_4x4
+    bne        .loop_4x\h
 
     pop         {pc}
     .ltorg
-- 
2.25.1

